import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output

from ts_app.dash_app import app
from ts_app.dashboard_components import (
    glossary,
    home_page,
    modelling,
    resources,
    sample,
    upload,
)

# Use a custom HTML template
app.index_string = resources.template

server = app.server


app.layout = html.Div(
    # Display current page
    [dcc.Location(id="url", refresh=False), html.Div(id="page-content")]
)


@app.callback(Output("page-content", "children"), [Input("url", "pathname")])
def render_page(pathname):
    """Display the page at the given pathname.

    Parameters
    ----------
    path_name : str
        The url to a page or dashboard layout.
    """
    if pathname == "/upload":
        # Get the dashboard layout with input from uploaded files
        data_source = upload.input_layout
    elif pathname == "/sample":
        # Get the dashboard layout which generates sample input data
        data_source = sample.input_layout
    elif pathname == "/glossary":
        # Display the glossary page
        return glossary.layout
    else:
        # Display the home page
        return home_page.layout

    # If the pathname is "/upload" or "/sample", create the appropriate
    # dashboard.
    return html.Div(
        id="dashboard",
        className="dashboard",
        children=[
            # Sidebar with data & model input
            html.Div(
                className="sidebar",
                children=[
                    # `data_source`: dropdowns to create a sample, or a file
                    # upload button.
                    html.Div(data_source, id="data-source"),
                    # Container for sample data
                    dcc.Store(id="sample-data-store"),
                    # Container for uploaded data
                    dcc.Store(id="upload-data-store"),
                    # Dropdowns for setting model parameters
                    modelling.param_input,
                ],
            ),
            html.Div(
                [
                    # A line graph of the fitted model, with predictions and
                    # forecast.
                    modelling.graph,
                    # Paragraphs with time series information
                    dcc.Markdown(resources.ts_details),
                    # Footer links to "home" and "glossary"
                    html.Footer(
                        [
                            html.A(
                                "Back to home",
                                href="/",
                                className="hvr-bob button",
                            ),
                            html.A(
                                "Browse glossary",
                                href="/glossary",
                                className="hvr-bob button",
                            ),
                        ]
                    ),
                ]
            ),
        ],
    )


if __name__ == "__main__":
    app.run_server(debug=True)



import pandas as pd
from dateutil.parser import ParserError as dtParserError


def process_upload(data):
    """Validate the data from the uploaded file.

    Parameters
    ----------
    data : pandas.Dataframe
        The data extracted from the file.

    Returns
    -------
    The error found, if any, or a pandas Series with the data's last column.
    """
    # Ensure the data has more than 30 values.
    if (n := len(data)) < 30:
        return f"""Please try again... the uploaded file has only {n} values,
                   but the minimum is set at 30."""

    # Ensure the dates(index) can be converted to datetime objects having a
    # consistent frequency.
    try:
        data.index = pd.to_datetime(data.index)
        # If date frequency can't be inferred, or is not consistent, some
        # statsmodels functions tend to break.
        if pd.infer_freq(data.index) in {None, "N"}:
            return """Please try again... a uniform date frequency (which is
                      needed in some of the time series functions used) could
                      not be determined."""
    except (dtParserError, ValueError):
        return """Please try again... it seems that the values in the 1st
                  column could not be read as dates."""

    # Ensure the data has at least 2 columns.
    # Since the first column is read as the index, there has to be at least
    # one other column with values.
    # If multiple columns exist, only the last is considered.
    try:
        data = data.iloc[:, -1]  # select the last column for analysis
    except IndexError:
        return """Please ensure that the data has a date index and at least
                  one other column with numeric values."""

    # Ensure the data values are numeric
    try:
        data = data.astype("float32")
    except ValueError:
        return """Please try again... it seems the values to be processed
                  could not be read as numbers."""



import dash_core_components as dcc
import dash_html_components as html
import pandas as pd
import plotly.graph_objs as go
from dash.dependencies import Input, Output
from ts_app.dash_app import app
from ts_app.ts_functions import create_arma_sample, fit_arima_model

param_input = html.Div(
    id="model-params",
    children=[
        html.H3("Model parameters"),
        # AR order dropdown
        html.P("AR Order"),
        dcc.Dropdown(
            id="model-ar",
            clearable=False,
            placeholder="AR order",
            searchable=False,
            value=1,
            options=[{"label": f"{i}", "value": i} for i in range(1, 5)],
        ),
        # Differencing order dropdown
        html.P("Differencing Order"),
        dcc.Dropdown(
            id="model-diff",
            clearable=False,
            value=1,
            placeholder="Differencing",
            searchable=False,
            options=[{"label": f"{i}", "value": i} for i in range(1, 5)],
        ),
        # MA order dropdown
        html.P("MA Order"),
        dcc.Dropdown(
            id="model-ma",
            clearable=False,
            placeholder="MA order",
            searchable=False,
            value=1,
            options=[{"label": f"{i}", "value": i} for i in range(1, 5)],
        ),
    ],
)

graph = html.Div(
    className="graph-container",
    children=[
        dcc.Loading(
            color="#777",
            children=[
                dcc.Graph(id="time-series-graph"),
            ],
        )
    ],
)


@app.callback(
    Output("time-series-graph", "figure"),
    [
        Input("model-ar", "value"),
        Input("model-diff", "value"),
        Input("model-ma", "value"),
        Input("url", "pathname"),
        Input("sample-data-store", "data"),
        Input("upload-data-store", "data"),
    ],
)
def model_and_predict(
    ar_order, diff_order, ma_order, input_source, sample, upload
):
    """Fit an ARIMA model each time model parameters or input data are
    modified, then plot the results.

    Parameters
    ----------
    ar_order, ma_order, diff_order : int
        The AR order, Differencing order & MA order for the ARIMA model.
    input_souce : {"/upload", "/sample"}
        The data source.
    sample, upload : dict
        Stored sample-data or uploaded-file-data respectively.
    """
    if input_source == "/upload":
        data_store = upload
    elif input_source == "/sample":
        data_store = sample

    if data_store is None:  # Default if there's no data
        # Create a random sample
        filename = "a random sample"
        data = create_arma_sample()
    else:
        filename = data_store["filename"]
        json_data = data_store["data"]
        data = pd.read_json(json_data, orient="index", typ="series")

    # Fit an ARIMA model on the obtained data, and get predictions
    predictions, forecast = fit_arima_model(
        data, ar_order, diff_order, ma_order
    )

    # Plot a line graph of the predictions
    fig = go.Figure()
    fig.add_trace(
        go.Scatter(y=data, x=data.index, mode="lines", name="actual data")
    )
    fig.add_trace(
        go.Scatter(
            x=predictions.index,
            y=predictions,
            mode="lines",
            name="predictions",
        )
    )
    fig.add_trace(
        go.Scatter(x=forecast.index, y=forecast, mode="lines", name="forecast")
    )

    model_info = f"ARIMA({ar_order}, {diff_order}, {ma_order})"

    fig.update_layout(
        paper_bgcolor="#eee",
        plot_bgcolor="#eee",
        title=f"An {model_info} model fitted on {filename}",
    )
    return fig



import json
from pkgutil import get_data

import dash_html_components as html

# Get glossary data from the ts_app package as bytes
glossary_data = get_data("ts_app", "assets/glossary.json")

# Get the glossary data as a json object
definitions = json.loads(glossary_data.decode())


components = sum(  # Concatenate the list of title-definition pairs
    [
        [
            html.Dt(item["title"], id=item["title"]),
            html.Dd(item["definition"], id=item["definition"]),
        ]
        for item in definitions
    ],
    start=[],
)

layout = html.Div(
    [
        html.H2("Glossary of Time Series Terms"),
        html.Dl(components),  # Descriptions list
        html.Footer(
            [html.A("Back to home", href="/", className="hvr-bob button")]
        ),
    ]
)



from datetime import date

import numpy as np
import pandas as pd
from statsmodels.api import tsa


def create_arma_sample(ar_order=1, ma_order=1, size=100):
    """Get a random ARMA sample.

    Parameters
    ----------
    ar_order, ma_order, size : int
        Values for the desired AR order, MA order and sample size.

    Returns
    -------
    An ARMA sample as a pandas Series.
    """
    ar_coeff = np.linspace(1, -0.9, ar_order + 1)  # arbitrary ar coefficients
    ma_coeff = np.linspace(1, 0.9, ma_order + 1)  # arbitrary ma coefficients
    sample = tsa.ArmaProcess(ar_coeff, ma_coeff).generate_sample(size)
    index = pd.date_range(start=date.today(), periods=size, freq="D")

    return pd.Series(sample, index=index, name="sample")


def fit_arima_model(data, ar_order=1, diff=0, ma_order=1):
    """Fit an ARIMA model on the data and get predictions.

    Parameters
    ----------
    data : pandas.Series
        The data to model, indexed by date.
    ar_order, ma_order, diff : int
        Values for the AR order, MA order and degree of Differencing
        respectively.

    Returns
    -------
    In-sample predictions covering the latter 30% of the data, and a 14-period
    out-of-sample forecast.
    """
    # Fit an ARIMA model
    arima_model = tsa.arima.ARIMA(data, order=(ar_order, diff, ma_order)).fit()
    # Get in-sample predictions
    n = len(data)
    predictions = arima_model.predict(start=int(0.7 * n), end=n)
    # Get forecast
    forecast = arima_model.predict(start=n, end=n + 14)

    return predictions, forecast



ts_details = """
Graphs help **visualise** the [trend](/glossary#Trend) in the data. They
clearly reveal whether there's been an increase, decrease or no change in the
values over time. Graphs are also useful in assessing the  **goodness of fit**.
In general, a good model should reasonably replicate the behaviour of the data.

Additionally, graphs help discover [seasonal](/glossary#Seasonality) and
[cyclic patterns](/glossary#Cyclic%20patterns) present in the data. These
usually manifest as occasional peaks or troughs. [Autocorrelation][1] and
[Partial-Autocorrelation][2] plots can provide hints on a potentially suitable
model to start with. [This article][3] describes how.

Testing for [stationarity][4], and filtering out seasonal & trend effects is an
essential first step. Model fitting usually requires that the time series data
be stationary, and mayfrom datetime import date even contain transformations
to make the data stationary.

[1]: https://en.wikipedia.org/wiki/Autocorrelation
[2]: https://en.wikipedia.org/wiki/Partial_autocorrelation_function
[3]: https://en.wikipedia.org/wiki/Box%E2%80%93Jenkins_method#Autocorrelation_\
and_partial_autocorrelation_plots
[4]: https://cran.r-project.org/web/packages/TSTutorial/vignettes/Stationary\
.pdf
"""

template = """
<!DOCTYPE html>
<html>

<head>
    {%metas%}
    <meta name="viewport" content="width=device-width, initial-scale=1">
    <title>{%title%}</title>
    {%favicon%}
    {%css%}
    <style>
        body {
            background-image: url("/assets/background_large.jpg");
            margin: 5% 10%;
            font: oblique 1.25em Georgia, serif;
        }

        @media only screen and (max-width: 640px) {
            body {
                background-image: url("/assets/background_small.jpg");
                margin: 10%;
            }
        }
    </style>
</head>

<body>
    {%app_entry%}
    <footer>
        {%config%}
        {%scripts%}
        {%renderer%}
    </footer>
</body>

</html>
"""



import logging
import webbrowser

from waitress import serve

from ts_app.dashboard import server

# Only log errors. `ConvergenceWarning`s and `ValueWarning`s are all too
# frequent when fitting the statsmodels ARIMA model on arbitrary data.
logging.basicConfig(level=logging.ERROR)
logging.captureWarnings(True)


__version__ = '0.1.1'


def run_app():
    """Start the app server, and launch a browser to view it."""

    # Open a new window of the default browser and point it to the where the
    # app is running.
    webbrowser.open('http://localhost:8000', new=1)

    # Serve the app using waitress
    print('Starting server at http://localhost:8000 ...')
    serve(server, host='localhost', port=8000)



import dash_core_components as dcc
import dash_html_components as html

layout = html.Div(
    [
        # Introduction
        dcc.Markdown(
            """
# Time Series App

---

A simple app to learn about, and apply [time series analysis][1] techniques. A
[glossary](/glossary) of terms is available to help get familiar with the main
concepts.

In a nutshell, Time Series Analysis purposes to learn and emulate the behaviour
of data over time. Time Series Forecasting then leverages this knowledge to
estimate future values.

![An example of time series analysis results, plotted](/assets/ts.svg)

To give it a try, you can upload a file or create a sample to analyse:

[1]: https://en.wikipedia.org/wiki/Time_series
"""
        ),
        html.A("Upload a file", href="/upload", className="hvr-bob button"),
        html.A("Process a sample", href="/sample", className="hvr-bob button"),
    ]
)



from base64 import b64decode
from io import BytesIO, StringIO

import dash_core_components as dcc
import dash_html_components as html
import pandas as pd
from dash.dependencies import Input, Output
from ts_app.dash_app import app
from ts_app.file_upload import process_upload

input_layout = html.Div(
    [
        # File upload button
        dcc.Upload(
            id="upload-data",
            accept=".csv,.xls,.xlsx",
            min_size=10,
            children=[
                html.Button("Upload a file", className="hvr-bob button"),
            ],
        ),
        # Container for the filename or error message
        html.P(id="file-info"),
    ]
)


@app.callback(
    [
        Output("file-info", "children"),
        Output("file-info", "style"),
        Output("upload-data-store", "data"),
    ],
    [Input("upload-data", "contents"), Input("upload-data", "filename")],
)
def upload_file(contents, filename):
    """Extract, validate and process data from uploaded files.

    parameters
    ----------
    contents : str
        Base64-encoded string with the file's contents
    filename : str
        The name of the uploaded file

    Returns
    -------
    A tuple - (file_info message, file_info style, data to store).
    """
    if contents is None:
        # If no file has been uploaded
        return (
            "",  # No file information
            {},  # No special style
            None,  # No data to store
        )
    else:
        content_type, content_string = contents.split(",")

    # Decode file content
    file_content = b64decode(content_string)

    # Pack the contents in a DataFrame
    try:
        if ".csv" in filename:
            # If the user uploaded a CSV file
            df = pd.read_csv(
                StringIO(file_content.decode("utf-8")), index_col=0
            )
        elif ".xls" in filename:
            # If the user uploaded an excel file
            df = pd.read_excel(BytesIO(file_content), index_col=0)
    except Exception as e:
        print(e)
        return (
            html.Div(["There was an error processing the file."]),
            {"color": "orangered"},  # Error-message text color
            None,  # No data to store
        )

    # Process the parsed data
    if (error := process_upload(data=df)) is not None:
        # If an error message is returned
        return (
            error,  # Error message
            {"color": "orangered"},  # Error-message text color
            None,  # No data to store
        )
    else:
        # If file-upload and data-extraction succeed
        return (
            f"Analysing {filename}",  # Uploaded file's name
            {"color": "#31bf2c"},  # Successful-upload text color
            {"filename": filename, "data": df.iloc[:, -1].to_json()},
        )
